==> palantir-2.6.tgz: palantir-2.6/server/audio.c <==

/*
    Audio handling functions.

    Copyright 2000-2003 by David Santinoli <david@santinoli.com>


GSM implementation notes
------------------------

Buffers used for I/O to /dev/dsp are arrays of SHORT, so their size in bytes is
2 * (array depth). That's why we see, say, 320 instead of 160 in many places.

Microsoft's WAV49 format requires encoded packets to be alternately 33- and
32-byte (start with 32 when encoding, 33 when decoding).
Both translate to 160 audio samples.
*/

#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <pthread.h>
#include <signal.h>
#include "libgsm/inc/gsm.h"
#include "audio.h"
#include "diag.h"

#define WAV49 1
#define GSM_SAMPLE_RATE 8000
#define REC_LEVEL   80

static const char rcsid[]="$Id: audio.c,v 1.15 2005/08/16 00:05:13 santinod Exp $";

/* Since 'gsm's are actually pointers, they are init'ed to NULL, so the
   gsm_destroy()s (in audio_cleanup) do no harm if the first gsm_create fails.
*/
static gsm gsm_tx_handle, gsm_rx_handle;
gsm_byte gsm_output_buf[66];
static gsm_signal gsm_input_buf[2*160]; /* we encode 2 packets per write */
static int audio_fd=-1;
static int audio_http_fd; /* obtained from accept() */
static pthread_t audio_server_tid, audio_client_tid;
static int downsample=0;
static short int *audio_read_buf=0;
enum audio_status aud_status;

int audio_init(char *device, int setup_mixer)
{
 int fd_flags;
 audio_buf_info auinfo;
 int format, stereo=0, speed=GSM_SAMPLE_RATE,
     frag_size=0x00020008; /* use 256-byte fragments */
 int wav49=1;

 aud_status=A_MUTE;

 /* audio dev setup */

 if((audio_fd=open(device,O_RDWR|O_NONBLOCK,0))==-1) {
    diag(SS_AUDIO|ERROR,"Can't open audio device %s\n",device);
    return 1;
    }

 fd_flags=fcntl(audio_fd,F_GETFL);
 if(fd_flags==-1) {
    diag(SS_AUDIO|ERROR,"Can't get audio fd flags\n");
    return 1;
    }
 if(fcntl(audio_fd,F_SETFL,fd_flags&~O_NONBLOCK)==-1) {
    diag(SS_AUDIO|DIERR,"Can't set audio fd flags\n");
    return 1;
    }

 if(ioctl(audio_fd,SNDCTL_DSP_SETDUPLEX,0)==-1)
    diag(SS_AUDIO|DEBUG,"Audio device supports half duplex only\n");
 else
    diag(SS_AUDIO|DEBUG,"Audio device supports full duplex\n");

 format=AFMT_S16_LE;
 if(ioctl(audio_fd,SNDCTL_DSP_SETFMT,&format)==-1) {
    diag(SS_AUDIO|ERROR,"ioctl(SNDCTL_DSP_SETFMT)\n");
    return 1;
    }
 if(format!=AFMT_S16_LE) {
    diag(SS_AUDIO|ERROR,"Can't set audio sample format\n");
    return 1;
    }

 if(ioctl(audio_fd,SNDCTL_DSP_STEREO,&stereo)==-1) {
    diag(SS_AUDIO|ERROR,"ioctl(SNDCTL_DSP_STEREO)\n");
    return 1;
    }

 if(stereo!=0) {
    diag(SS_AUDIO|ERROR,"Can't set mono channel mode\n");
    return 1;
    }

 if(ioctl(audio_fd,SNDCTL_DSP_SPEED,&speed)==-1) {
    diag(SS_AUDIO|ERROR,"ioctl(SNDCTL_DSP_SPEED)\n");
    return 1;
    }
 diag(SS_AUDIO|DEBUG,"Audio sample rate set to %d Hz\n",speed);
 if(speed!=GSM_SAMPLE_RATE) {
    if(speed>0 && speed%GSM_SAMPLE_RATE==0) {
       diag(SS_AUDIO|INFO,"Will do software downsampling to 8000 Hz\n");
       downsample=speed/GSM_SAMPLE_RATE;
       }
    else {
       diag(SS_AUDIO|ERROR,"Unable to set sample rate to %d Hz\n",GSM_SAMPLE_RATE);
       return 1;
       }
    }

 if(ioctl(audio_fd,SNDCTL_DSP_SETFRAGMENT,&frag_size)==-1) {
    diag(SS_AUDIO|ERR,"ioctl(SNDCTL_DSP_SETFRAGMENT)");
    /* return 1; */
    }

 if(ioctl(audio_fd,SNDCTL_DSP_GETISPACE,&auinfo)==-1) {
    diag(SS_AUDIO|ERR,"Cannot get audio input buffering info");
    /* return 1; */
    }
 else diag(SS_AUDIO|DEBUG,"Audio input buffering: %d fragments of %d bytes each\n",auinfo.fragstotal, auinfo.fragsize);

 if(ioctl(audio_fd,SNDCTL_DSP_GETOSPACE,&auinfo)==-1) {
    diag(SS_AUDIO|ERR,"Cannot get audio output buffering info");
    /* return 1; */
    }
 else diag(SS_AUDIO|DEBUG,"Audio output buffering: %d fragments of %d bytes each\n",auinfo.fragstotal, auinfo.fragsize);

 /* mixer setup */
 /* FIX: Some Creative cards disregard proper input channel volume setting;
    true input level is called "Input Gain" and is a separate slider */

 if(setup_mixer) {
    int mixer_fd, bits, volume, msuccess=0;
    if((mixer_fd=open("/dev/mixer",O_RDWR,0))==-1) {
       diag(SS_AUDIO|ERROR,"Can't open mixer device\n");
       return 1;
       }

    if(ioctl(mixer_fd,SOUND_MIXER_READ_RECMASK,&bits)==-1) {
       diag(SS_AUDIO|ERROR,"ioctl(SOUND_MIXER_READ_RECMASK)\n");
       }
    else if(!(bits&(1<<SOUND_MIXER_MIC))) {
       diag(SS_AUDIO|ERROR,"No microphone input\n");
       }
    else {
       /* mic or line? */
       bits=(1<<SOUND_MIXER_MIC);
       if(ioctl(mixer_fd,SOUND_MIXER_WRITE_RECSRC,&bits)==-1) {
          diag(SS_AUDIO|ERROR,"Can't set audio input source\n");
	  }
       else {
          volume=(REC_LEVEL<<8)|REC_LEVEL;
          if(ioctl(mixer_fd,SOUND_MIXER_WRITE_MIC,&volume)==-1)
             diag(SS_AUDIO|ERROR,"ioctl(SOUND_MIXER_WRITE_MIC)\n");
	  else msuccess=1; 
	  }
       }

    close(mixer_fd);
    if(!msuccess) return 1;
    }

 if(!(gsm_tx_handle=gsm_create()) || !(gsm_rx_handle=gsm_create())) {
    diag(SS_AUDIO|ERROR,"Error creating GSM handles\n");
    return 1;
    }

#if WAV49
 if(gsm_option(gsm_tx_handle,GSM_OPT_WAV49,&wav49)==-1 ||
       gsm_option(gsm_rx_handle,GSM_OPT_WAV49,&wav49)==-1) {
    diag(SS_AUDIO|ERROR,"Cannot set WAV49 option in GSM codec (audio communication with Windows clients will be corrupted)\n");
    return 1;
    }
#endif

 if(downsample)
    if(!(audio_read_buf=malloc(downsample*sizeof(gsm_input_buf)))) {
       diag(SS_AUDIO|ERROR,"Cannot allocate audio downsampling buffer\n");
       return 1;
       }

 return 0;
}


void audio_cleanup(void)
{
 close(audio_fd);
 gsm_destroy(gsm_tx_handle);
 gsm_destroy(gsm_rx_handle);
}


void audio_launch_threads(int sock, struct sockaddr_in *name)
{
 pthread_attr_t thr_attr;
 sigset_t sigs;
 int launch_s, launch_c;
 
 aud_status=A_CURRENT;
 audio_http_fd=sock;
/*
 ling.l_onoff=0;
 setsockopt(audio_http_fd,SOL_SOCKET,SO_LINGER,&ling,sizeof(ling));
*/
 sigemptyset(&sigs);
 sigaddset(&sigs,SIGHUP);
 pthread_sigmask(SIG_BLOCK,&sigs,0);
 pthread_attr_init(&thr_attr);
 pthread_attr_setdetachstate(&thr_attr,PTHREAD_CREATE_JOINABLE);
 launch_s=pthread_create(&audio_server_tid,&thr_attr,(void*(*)(void*))audio_server,NULL);
 if(launch_s)
    diag(SS_AUDIO|ERR,"unable to create audio server thread");
 pthread_attr_init(&thr_attr);
 pthread_attr_setdetachstate(&thr_attr,PTHREAD_CREATE_JOINABLE);
 launch_c=pthread_create(&audio_client_tid,&thr_attr,(void*(*)(void*))audio_client,NULL);
 if(launch_c)
    diag(SS_AUDIO|ERR,"unable to create audio client thread");
 pthread_sigmask(SIG_UNBLOCK,&sigs,0);

 if(!launch_s || !launch_c)
    diag(SS_AUDIO|DEBUG,"%s:%hu: audio thread(s) started\n",
       inet_ntoa(name->sin_addr),ntohs(name->sin_port));
 else {
    close(sock);
    aud_status=A_MUTE;
    }
}


void audio_terminate_threads(void)
{
 int e;

 diag(SS_AUDIO|DEBUG,"attempting to terminate audio threads...\n");

 pthread_cancel(audio_server_tid);
 pthread_cancel(audio_client_tid);

 if( (e=pthread_join(audio_server_tid,NULL)) )
   if(e!=ESRCH) diag(SS_AUDIO|ERROR,"error joining audio_server thread");
 if( (e=pthread_join(audio_client_tid,NULL)) )
   if(e!=ESRCH) diag(SS_AUDIO|ERROR,"error joining audio_client thread");
 close(audio_http_fd);
 aud_status=A_MUTE;
 diag(SS_AUDIO|DEBUG,"audio threads terminated.\n");
 /* diag(SS_AUDIO|DEBUG,"frame index: %d\n",
       gsm_option(gsm_rx_handle,GSM_OPT_FRAME_INDEX,NULL)); */

}


/* sound card -> remote client */
void *audio_server(void)
{
 int bytes=0, stale, n, gsmframe=0;
 /* int packets=0; */
 float time;
 struct timeval tv1,tv2;
 audio_buf_info auinfo;
 char scrap[10000];

#if WAV49
 if(gsm_option(gsm_tx_handle,GSM_OPT_FRAME_INDEX,&gsmframe)==-1)
    diag(SS_AUDIO|DIE,"Error resetting tx GSM frame index\n");
#endif

 if(gettimeofday(&tv1,NULL)) diag(SS_AUDIO|ERR,"audio_server: gettimeofday()");

 /* flush read buffer. Especially useful with large input buffers (i.e.
    many fragments) */
 if(ioctl(audio_fd,SNDCTL_DSP_GETISPACE,&auinfo)==0) {
    diag(SS_AUDIO|DEBUG,"Audio input prebuffered %d bytes\n",auinfo.bytes);
    stale=auinfo.bytes;
    while(stale>0) {
       //diag(SS_AUDIO|DEBUG,"Dequeuing %d bytes (%d)\n",stale, stale);
       n=read(audio_fd,scrap,stale>10000?10000:stale);
       if(n==-1) {
          diag(SS_AUDIO|ERR,"Audio stream read error\n");
          goto out;
          }
       stale-=n;
       }
    }

 while(1) {
    if(!downsample)
       n=read(audio_fd,(short int*)gsm_input_buf,sizeof(gsm_input_buf));
    else {
       int o1,o2;
       n=read(audio_fd,(short int*)audio_read_buf,downsample*sizeof(gsm_input_buf));
       for(o1=o2=0;o1<downsample*320;o1+=downsample,o2++)
          gsm_input_buf[o2]=audio_read_buf[o1];
       n/=downsample;
       }
    if(n!=sizeof(gsm_input_buf)) {
       diag(SS_AUDIO|ERR,"Audio stream read error");
       break;
       }
    gsm_encode(gsm_tx_handle,gsm_input_buf,gsm_output_buf);
    gsm_encode(gsm_tx_handle,gsm_input_buf+160,gsm_output_buf+33-WAV49);
    /* diag(SS_AUDIO|NOTIME|DEBUG,"."); */

#if 0
    /* dropping one packet out of N improves latency but introduces clicks */
    packets++;
    if(packets==25) {
       packets=0;
       continue;
       }
#endif

    if(write(audio_http_fd,gsm_output_buf,66-WAV49)!=66-WAV49) {
       if(errno!=EPIPE) diag(SS_AUDIO|ERR,"audio write");
       break;
       }
    bytes+=66-WAV49;

#if 0
    if(cops->timeout) {
       if(gettimeofday(&tv2,NULL)) diag(SS_AUDIO|ERR,"audio_server: gettimeofday()");
       if(tv2.tv_sec-tv1.tv_sec>cops->timeout) break;
       }
#endif
    } /* forever */

out:
 if(gettimeofday(&tv2,NULL)) diag(SS_AUDIO|ERR,"audio_server: gettimeofday()");
 //close(cops->fd);
 time=tv2.tv_sec-tv1.tv_sec+(tv2.tv_usec-tv1.tv_usec)/1000000.0;
 diag(SS_AUDIO|DEBUG,"audio server: wrote %d bytes in %.2f s - %.2f B/s\n",
    bytes,time,bytes/time);

 return NULL;
}


/* remote client -> sound card */
void *audio_client(void)
{
 gsm_frame gsm_rx_input_buf;
 gsm_signal gsm_rx_output_buf[160];
 int pos,n, gsmframe=0;
 int frame_size=33;

#if WAV49
 if(gsm_option(gsm_rx_handle,GSM_OPT_FRAME_INDEX,&gsmframe)==-1)
    diag(SS_AUDIO|DIE,"Error resetting rx GSM frame index\n");
#endif

 while(1) {
    pos=0;
    do {
       n=read(audio_http_fd, gsm_rx_input_buf+pos, frame_size-pos);
       if(n==-1) {
          diag(SS_AUDIO|ERR,"Error receiving data");
          goto out;
          }
       else if(n==0) {
          diag(SS_AUDIO|DEBUG,"End of stream\n");
          goto out;
          }
       pos+=n;
       /*diag(SS_AUDIO|DEBUG,"Received packet (%d bytes).\n",n);*/
       } while(pos<frame_size);
    if (gsm_decode(gsm_rx_handle, gsm_rx_input_buf, gsm_rx_output_buf) < 0)
       diag(SS_AUDIO|ERR,"Frame decode failed\n");
    if (write(audio_fd, gsm_rx_output_buf, 320) != 320)
       diag(SS_AUDIO|ERR,"Error playing audio\n");

#if WAV49
    frame_size=65-frame_size;
#endif
    }
out:
 return NULL;
}

==> palantir-2.7.tgz: palantir-2.7/server/audio.c <==

/*
    Audio handling functions.

    Copyright 2000-2007 by David Santinoli <david@santinoli.com>


GSM implementation notes
------------------------

Buffers used for I/O to audio device are arrays of SHORT, so their size in
bytes is 2 * (array depth). That's why we see, say, 320 instead of 160 in many
places.

Microsoft's WAV49 format requires encoded packets to be alternately 33- and
32-byte (start with 32 when encoding, 33 when decoding).
Both translate to 160 audio samples.
*/

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>
#include <pthread.h>
#include <signal.h>
#include "libgsm/inc/gsm.h"
#include "audio.h"
#include "diag.h"

#define WAV49 1
#define GSM_SAMPLE_RATE 8000
#define REC_LEVEL   80

static const char rcsid[]="$Id: audio.c,v 1.16 2007/05/18 22:05:56 santinod Exp $";

/* Since 'gsm's are actually pointers, they are init'ed to NULL, so the
   gsm_destroy()s (in audio_cleanup) do no harm if the first gsm_create fails.
*/
static gsm gsm_tx_handle, gsm_rx_handle;
gsm_byte gsm_output_buf[66];
static gsm_signal gsm_input_buf[2*160]; /* we encode 2 packets per write */
static int audio_in_fd=-1, audio_out_fd=-1;
static int audio_http_fd; /* obtained from accept() */
static pthread_t audio_server_tid, audio_client_tid;
static int downsample=0;
static short int *audio_read_buf=0;
enum audio_status aud_status;

int audio_channel_init(char *device, enum audio_mode *dir, int *fd)
{
 int fd_flags;
 audio_buf_info auinfo;
 int format, stereo=0, speed=GSM_SAMPLE_RATE,
     frag_size=0x00020008; /* use 256-byte fragments */

 diag(SS_AUDIO|DEBUG,"Setting up audio device %s\n",device);

 if(*dir==A_MODE_DUPLEX) {
    if((*fd=open(device,O_RDWR|O_NONBLOCK,0))==-1) {
       diag(SS_AUDIO|ERROR,"Can't open audio device %s in duplex mode. Reverting to input only\n",device);
       *dir=A_MODE_IN;
       }
    if(*fd>=0 && ioctl(*fd,SNDCTL_DSP_SETDUPLEX,0)==-1) {
       diag(SS_AUDIO|DEBUG,"Audio device supports half duplex only\n");
       close(*fd);
       *dir=A_MODE_IN;
       }
    else
       diag(SS_AUDIO|DEBUG,"Audio device supports full duplex\n");
    }

 if(*dir==A_MODE_IN) {
    if((*fd=open(device,O_RDONLY|O_NONBLOCK,0))==-1) {
       diag(SS_AUDIO|ERR,"Can't open audio device %s in input mode",device);
       return 1;
       }
    }

 if(*dir==A_MODE_OUT) {
    if((*fd=open(device,O_WRONLY|O_NONBLOCK,0))==-1) {
       diag(SS_AUDIO|ERR,"Can't open audio device %s in output mode",device);
       return 1;
       }
    }

 fd_flags=fcntl(*fd,F_GETFL);
 if(fd_flags==-1) {
    diag(SS_AUDIO|ERROR,"Can't get audio fd flags\n");
    return 1;
    }
 if(fcntl(*fd,F_SETFL,fd_flags&~O_NONBLOCK)==-1) {
    diag(SS_AUDIO|DIERR,"Can't set audio fd flags\n");
    return 1;
    }

 format=AFMT_S16_LE;
 if(ioctl(*fd,SNDCTL_DSP_SETFMT,&format)==-1) {
    diag(SS_AUDIO|ERROR,"ioctl(SNDCTL_DSP_SETFMT)\n");
    return 1;
    }
 if(format!=AFMT_S16_LE) {
    diag(SS_AUDIO|ERROR,"Can't set audio sample format\n");
    return 1;
    }

 if(ioctl(*fd,SNDCTL_DSP_STEREO,&stereo)==-1) {
    diag(SS_AUDIO|ERROR,"ioctl(SNDCTL_DSP_STEREO)\n");
    return 1;
    }

 if(stereo!=0) {
    diag(SS_AUDIO|ERROR,"Can't set mono channel mode\n");
    return 1;
    }

 if(ioctl(*fd,SNDCTL_DSP_SPEED,&speed)==-1) {
    diag(SS_AUDIO|ERROR,"ioctl(SNDCTL_DSP_SPEED)\n");
    return 1;
    }
 diag(SS_AUDIO|DEBUG,"Audio sample rate set to %d Hz\n",speed);
 if(speed!=GSM_SAMPLE_RATE) {
    if(*dir==A_MODE_IN && speed>0 && speed%GSM_SAMPLE_RATE==0) {
       diag(SS_AUDIO|INFO,"(Audio input will be downsampled to 8000 Hz in software)\n");
       downsample=speed/GSM_SAMPLE_RATE;
       }
    else {
       diag(SS_AUDIO|ERROR,"Unable to set sample rate to %d Hz\n",GSM_SAMPLE_RATE);
       return 1;
       }
    }

 if(ioctl(*fd,SNDCTL_DSP_SETFRAGMENT,&frag_size)==-1) {
    diag(SS_AUDIO|ERR,"ioctl(SNDCTL_DSP_SETFRAGMENT)");
    /* return 1; */
    }

 if(*dir==A_MODE_IN || *dir==A_MODE_DUPLEX) {
    if(ioctl(*fd,SNDCTL_DSP_GETISPACE,&auinfo)==-1) {
       diag(SS_AUDIO|ERR,"Cannot get audio input buffering info");
       /* return 1; */
       }
    else diag(SS_AUDIO|DEBUG,"Audio input buffering: %d fragments of %d bytes each\n",auinfo.fragstotal, auinfo.fragsize);
    }

 if(*dir==A_MODE_OUT || *dir==A_MODE_DUPLEX) {
    if(ioctl(*fd,SNDCTL_DSP_GETOSPACE,&auinfo)==-1) {
       diag(SS_AUDIO|ERR,"Cannot get audio output buffering info");
       /* return 1; */
       }
    else diag(SS_AUDIO|DEBUG,"Audio output buffering: %d fragments of %d bytes each\n",auinfo.fragstotal, auinfo.fragsize);
    }
 return 0;
}


int audio_init(char *in_device, char *out_device, int setup_mixer)
{
 int wav49=1;
 enum audio_mode channel_mode=-1;

 aud_status=A_MUTE;

 if(in_device) {
    if(out_device && !strcmp(in_device,out_device))
       channel_mode=A_MODE_DUPLEX;
    else
       channel_mode=A_MODE_IN;
    if(audio_channel_init(in_device,&channel_mode,&audio_in_fd))
       return 1;
    if(channel_mode==A_MODE_DUPLEX) audio_out_fd=audio_in_fd;
    }

 if(out_device && channel_mode!=A_MODE_DUPLEX) {
    channel_mode=A_MODE_OUT;
    if(audio_channel_init(out_device,&channel_mode,&audio_out_fd))
       return 1;
    }

 /* mixer setup */
 /* FIX: Some Creative cards disregard proper input channel volume setting;
    true input level is called "Input Gain" and is a separate slider */

 if(setup_mixer) {
    int mixer_fd, bits, volume, msuccess=0;
    if((mixer_fd=open("/dev/mixer",O_RDWR,0))==-1) {
       diag(SS_AUDIO|ERROR,"Can't open mixer device\n");
       return 1;
       }

    if(ioctl(mixer_fd,SOUND_MIXER_READ_RECMASK,&bits)==-1) {
       diag(SS_AUDIO|ERROR,"ioctl(SOUND_MIXER_READ_RECMASK)\n");
       }
    else if(!(bits&(1<<SOUND_MIXER_MIC))) {
       diag(SS_AUDIO|ERROR,"No microphone input\n");
       }
    else {
       /* mic or line? */
       bits=(1<<SOUND_MIXER_MIC);
       if(ioctl(mixer_fd,SOUND_MIXER_WRITE_RECSRC,&bits)==-1) {
          diag(SS_AUDIO|ERROR,"Can't set audio input source\n");
	  }
       else {
          volume=(REC_LEVEL<<8)|REC_LEVEL;
          if(ioctl(mixer_fd,SOUND_MIXER_WRITE_MIC,&volume)==-1)
             diag(SS_AUDIO|ERROR,"ioctl(SOUND_MIXER_WRITE_MIC)\n");
	  else msuccess=1; 
	  }
       }

    close(mixer_fd);
    if(!msuccess) return 1;
    }

 if(!(gsm_tx_handle=gsm_create()) || !(gsm_rx_handle=gsm_create())) {
    diag(SS_AUDIO|ERROR,"Error creating GSM handles\n");
    return 1;
    }

#if WAV49
 if(gsm_option(gsm_tx_handle,GSM_OPT_WAV49,&wav49)==-1 ||
       gsm_option(gsm_rx_handle,GSM_OPT_WAV49,&wav49)==-1) {
    diag(SS_AUDIO|ERROR,"Cannot set WAV49 option in GSM codec (audio communication with Windows clients will be corrupted)\n");
    return 1;
    }
#endif

 if(downsample)
    if(!(audio_read_buf=malloc(downsample*sizeof(gsm_input_buf)))) {
       diag(SS_AUDIO|ERROR,"Cannot allocate audio downsampling buffer\n");
       return 1;
       }

 return 0;
}


void audio_cleanup(void)
{
 if(audio_in_fd>=0)
    close(audio_in_fd);
 if(audio_out_fd>=0 && audio_out_fd!=audio_in_fd)
    close(audio_out_fd);
 gsm_destroy(gsm_tx_handle);
 gsm_destroy(gsm_rx_handle);
}


void audio_launch_threads(int sock, struct sockaddr_storage *name)
{
 pthread_attr_t thr_attr;
 sigset_t sigs;
 int launch_s=0, launch_c=0;
 char srv_addr[NI_MAXHOST];

 aud_status=A_CURRENT;
 audio_http_fd=sock;
/*
 ling.l_onoff=0;
 setsockopt(audio_http_fd,SOL_SOCKET,SO_LINGER,&ling,sizeof(ling));
*/
 sigemptyset(&sigs);
 sigaddset(&sigs,SIGHUP);
 pthread_sigmask(SIG_BLOCK,&sigs,0);

 if(audio_in_fd>=0) {
    pthread_attr_init(&thr_attr);
    pthread_attr_setdetachstate(&thr_attr,PTHREAD_CREATE_JOINABLE);
    launch_s=pthread_create(&audio_server_tid,&thr_attr,(void*(*)(void*))audio_server,NULL);
    if(launch_s)
       diag(SS_AUDIO|ERR,"unable to create audio server thread");
    else
       diag(SS_AUDIO|DEBUG,"audio server thread started\n");
    }

 if(audio_out_fd>=0) {
    pthread_attr_init(&thr_attr);
    pthread_attr_setdetachstate(&thr_attr,PTHREAD_CREATE_JOINABLE);
    launch_c=pthread_create(&audio_client_tid,&thr_attr,(void*(*)(void*))audio_client,NULL);
    if(launch_c)
       diag(SS_AUDIO|ERR,"unable to create audio client thread");
    else
       diag(SS_AUDIO|DEBUG,"audio client thread started\n");
    }

 pthread_sigmask(SIG_UNBLOCK,&sigs,0);

 if(!launch_s || !launch_c) {
    if(getnameinfo((struct sockaddr *)name,sizeof(*name),
       srv_addr,sizeof(srv_addr),0,0,
       NI_NUMERICHOST|NI_NUMERICSERV)!=0) {
       strcpy(srv_addr,"unknown");
       }
    diag(SS_AUDIO|DEBUG,"%s - audio thread(s) started\n",srv_addr);
    }
 else {
    close(sock);
    aud_status=A_MUTE;
    }
}


void audio_terminate_threads(void)
{
 int e;

 diag(SS_AUDIO|DEBUG,"attempting to terminate audio threads...\n");

 if(audio_in_fd>=0)
    pthread_cancel(audio_server_tid);

 if(audio_out_fd>=0)
    pthread_cancel(audio_client_tid);

 if(audio_in_fd>=0 && (e=pthread_join(audio_server_tid,NULL)) )
   if(e!=ESRCH) diag(SS_AUDIO|ERROR,"error joining audio_server thread");
 if(audio_out_fd>=0 && (e=pthread_join(audio_client_tid,NULL)) )
   if(e!=ESRCH) diag(SS_AUDIO|ERROR,"error joining audio_client thread");

 close(audio_http_fd);
 aud_status=A_MUTE;
 diag(SS_AUDIO|DEBUG,"audio threads terminated.\n");
 /* diag(SS_AUDIO|DEBUG,"frame index: %d\n",
       gsm_option(gsm_rx_handle,GSM_OPT_FRAME_INDEX,NULL)); */

}


/* sound card -> remote client */
void *audio_server(void)
{
 int bytes=0, stale, n, gsmframe=0;
 /* int packets=0; */
 float time;
 struct timeval tv1,tv2;
 audio_buf_info auinfo;
 char scrap[10000];

#if WAV49
 if(gsm_option(gsm_tx_handle,GSM_OPT_FRAME_INDEX,&gsmframe)==-1)
    diag(SS_AUDIO|DIE,"Error resetting tx GSM frame index\n");
#endif

 if(gettimeofday(&tv1,NULL)) diag(SS_AUDIO|ERR,"audio_server: gettimeofday()");

 /* flush read buffer. Especially useful with large input buffers (i.e.
    many fragments) */
 if(ioctl(audio_in_fd,SNDCTL_DSP_GETISPACE,&auinfo)==0) {
    diag(SS_AUDIO|DEBUG,"Audio input prebuffered %d bytes\n",auinfo.bytes);
    /* as we deal with 16-bit samples, must discard an even number of bytes */
    stale=auinfo.bytes&(~1);
    while(stale>0) {
       n=read(audio_in_fd,scrap,stale>10000?10000:stale);
       diag(SS_AUDIO|DEBUG,"Discarded %d bytes (%d)\n",n, stale);
       if(n==-1) {
          diag(SS_AUDIO|ERR,"Audio stream read error\n");
          goto out;
          }
       stale-=n;
       }
    }

 while(1) {
    if(!downsample)
       n=read(audio_in_fd,(short int*)gsm_input_buf,sizeof(gsm_input_buf));
    else {
       int o1,o2;
       n=read(audio_in_fd,(short int*)audio_read_buf,downsample*sizeof(gsm_input_buf));
       for(o1=o2=0;o1<downsample*320;o1+=downsample,o2++)
          gsm_input_buf[o2]=audio_read_buf[o1];
       n/=downsample;
       }
    if(n!=sizeof(gsm_input_buf)) {
       diag(SS_AUDIO|ERR,"Audio stream read error");
       break;
       }
    gsm_encode(gsm_tx_handle,gsm_input_buf,gsm_output_buf);
    gsm_encode(gsm_tx_handle,gsm_input_buf+160,gsm_output_buf+33-WAV49);
    /* diag(SS_AUDIO|NOTIME|DEBUG,"."); */

#if 0
    /* dropping one packet out of N improves latency but introduces clicks */
    packets++;
    if(packets==25) {
       packets=0;
       continue;
       }
#endif

    if(write(audio_http_fd,gsm_output_buf,66-WAV49)!=66-WAV49) {
       if(errno!=EPIPE) diag(SS_AUDIO|ERR,"audio write");
       break;
       }
    bytes+=66-WAV49;

#if 0
    if(cops->timeout) {
       if(gettimeofday(&tv2,NULL)) diag(SS_AUDIO|ERR,"audio_server: gettimeofday()");
       if(tv2.tv_sec-tv1.tv_sec>cops->timeout) break;
       }
#endif
    } /* forever */

out:
 if(gettimeofday(&tv2,NULL)) diag(SS_AUDIO|ERR,"audio_server: gettimeofday()");
 //close(cops->fd);
 time=tv2.tv_sec-tv1.tv_sec+(tv2.tv_usec-tv1.tv_usec)/1000000.0;
 diag(SS_AUDIO|DEBUG,"audio server: wrote %d bytes in %.2f s - %.2f B/s\n",
    bytes,time,bytes/time);

 return NULL;
}


/* remote client -> sound card */
void *audio_client(void)
{
 gsm_frame gsm_rx_input_buf;
 gsm_signal gsm_rx_output_buf[160];
 int pos,n, gsmframe=0;
 int frame_size=33;

#if WAV49
 if(gsm_option(gsm_rx_handle,GSM_OPT_FRAME_INDEX,&gsmframe)==-1)
    diag(SS_AUDIO|DIE,"Error resetting rx GSM frame index\n");
#endif

 while(1) {
    pos=0;
    do {
       n=read(audio_http_fd, gsm_rx_input_buf+pos, frame_size-pos);
       if(n==-1) {
          diag(SS_AUDIO|ERR,"Error receiving data");
          goto out;
          }
       else if(n==0) {
          diag(SS_AUDIO|DEBUG,"End of stream\n");
          goto out;
          }
       pos+=n;
       /*diag(SS_AUDIO|DEBUG,"Received packet (%d bytes).\n",n);*/
       } while(pos<frame_size);
    if (gsm_decode(gsm_rx_handle, gsm_rx_input_buf, gsm_rx_output_buf) < 0)
       diag(SS_AUDIO|ERR,"Frame decode failed\n");
    if (write(audio_out_fd, gsm_rx_output_buf, 320) != 320)
       diag(SS_AUDIO|ERR,"Error playing audio\n");

#if WAV49
    frame_size=65-frame_size;
#endif
    }
out:
 return NULL;
}
